#include <nds.h>
#include "arm7.h"
#include "fw.h"
#include "ipc.h"
#include "../../arm9/include/wifi.h"


// Delay (in swiDelay units) for each bit transfer
#define RTC_DELAY 192

// Pin defines on RTC_CR
#define CS_0    (1<<6)
#define CS_1    ((1<<6) | (1<<2))
#define SCK_0   (1<<5)
#define SCK_1   ((1<<5) | (1<<1))
#define SIO_0   (1<<4)
#define SIO_1   ((1<<4) | (1<<0))
#define SIO_out (1<<4)
#define SIO_in  (1)

void SETRTC(uint8 * command)
{
	u32 bit;
	u8 data;
	u32 commandLength = 7;
	
	// Raise CS
	RTC_CR8 = CS_0 | SCK_1 | SIO_1;
	swiDelay(RTC_DELAY);
	RTC_CR8 = CS_1 | SCK_1 | SIO_1;
	swiDelay(RTC_DELAY);
	
	data = *command++;
	for (bit = 0; bit < 8; bit++) {
			RTC_CR8 = CS_1 | SCK_0 | SIO_out | (data>>7);
			swiDelay(RTC_DELAY);

			RTC_CR8 = CS_1 | SCK_1 | SIO_out | (data>>7);
			swiDelay(RTC_DELAY);

			data = data << 1;
		}
	

	// Write command bytes (high bit first)
	for ( ; commandLength > 0; commandLength--) {
		data = *command++;

		for (bit = 0; bit < 8; bit++) {
			RTC_CR8 = CS_1 | SCK_0 | SIO_out | (data & 1);
			swiDelay(RTC_DELAY);

			RTC_CR8 = CS_1 | SCK_1 | SIO_out | (data & 1);
			swiDelay(RTC_DELAY);

			data = data >> 1;
		}
	}
}

//---------------------------------------------------------------------------------
void startSound(int sampleRate, const void* data, u32 bytes, u8 channel, u8 vol,  u8 pan, u8 format) {
//---------------------------------------------------------------------------------
	SCHANNEL_TIMER(channel)  = SOUND_FREQ(sampleRate);
	SCHANNEL_SOURCE(channel) = (u32)data;
	SCHANNEL_LENGTH(channel) = bytes >> 2 ;
	SCHANNEL_CR(channel)     = SCHANNEL_ENABLE | SOUND_ONE_SHOT | SOUND_VOL(vol) | SOUND_PAN(pan) | (format==1?SOUND_8BIT:SOUND_16BIT);
}


//---------------------------------------------------------------------------------
s32 getFreeSoundChannel() {
//---------------------------------------------------------------------------------
	int i;
	for (i=0; i<16; i++) {
		if ( (SCHANNEL_CR(i) & SCHANNEL_ENABLE) == 0 ) return i;
	}
	return -1;
}

static bool touchInit = false;
static s32 xscale, yscale;
static s32 xoffset, yoffset;

void interrupthandler() {
	//u32 flags=REG_IF&REG_IE;
	
	static int heartbeat = 0;

	uint16 but=0, x=0, y=0, xpx=0, ypx=0, z1=0, z2=0, batt=0, aux=0;
	int t1=0, t2=0;
	uint32 temp=0;
	uint8 ct[sizeof(IPC->curtime)];
	u32 i;

	// Update the heartbeat
	heartbeat++;

	// Read the touch screen

	but = REG_KEYXY;

	if (!(but & (1<<6))) {

		touchPosition tempPos = touchReadXY();

		x = tempPos.x;
		y = tempPos.y;
		
		if ( !touchInit ) {
			xscale = ((PersonalData->calX2px - PersonalData->calX1px) << 19) / ((PersonalData->calX2) - (PersonalData->calX1));
			yscale = ((PersonalData->calY2px - PersonalData->calY1px) << 19) / ((PersonalData->calY2) - (PersonalData->calY1));
	
			xoffset = ((PersonalData->calX1 + PersonalData->calX2) * xscale  - ((PersonalData->calX1px + PersonalData->calX2px) << 19) ) / 2;
			yoffset = ((PersonalData->calY1 + PersonalData->calY2) * yscale  - ((PersonalData->calY1px + PersonalData->calY2px) << 19) ) / 2;
			touchInit = true;
		}
		//s16 px = ( touchPos.x * xscale - xoffset + xscale/2 ) >>19;
		//s16 py = ( touchPos.y * yscale - yoffset + yscale/2 ) >>19;
		//xpx = tempPos.px;
		//ypx = tempPos.py;
		
		s16 px = (x * xscale - xoffset + xscale/2 ) >> 19;
		s16 py = (y * yscale - yoffset + yscale/2 ) >> 19;
		if ( px < 0) px = 0;
		if ( py < 0) py = 0;
		if ( px > 255) px = 255;
		if ( py > 191) py = 191;
		xpx = px;
		ypx = py;
	}

	z1 = touchRead(TSC_MEASURE_Z1);
	z2 = touchRead(TSC_MEASURE_Z2);

	
	batt = touchRead(TSC_MEASURE_BATTERY);
	aux  = touchRead(TSC_MEASURE_AUX);

	// Read the time
	rtcGetTime((uint8 *)ct);
	i=ct[5]&0x40;		//Save the AM/PM flag for later use.
	ct[5]&=0x3F;		//Mask off the AM/PM flag, to prevent it
									//from affecting the BCDToInteger operation.
	BCDToInteger((uint8 *)&(ct[1]), 7);
	if((i==0x40) && (ct[5] < 12))	//If Flag indicates PM, and RTC is
																//in AM/PM mode, convert to 24 hour time.
		ct[5]+=12;

	// Read the temperature
	temp = touchReadTemperature(&t1, &t2);

	// Update the IPC struct
	IPC->heartbeat	= heartbeat;
	IPC->buttons		= but;
	IPC->touchX			= x;
	IPC->touchY			= y;
	IPC->touchXpx		= xpx;
	IPC->touchYpx		= ypx;
	IPC->touchZ1		= z1;
	IPC->touchZ2		= z2;
	IPC->battery		= batt;
	IPC->aux			= aux;
	
	for(i=0; i<sizeof(ct); i++) {
		IPC->curtime[i] = ct[i];
	}
	
	IPC->temperature = temp;
	IPC->tdiode1 = t1;
	IPC->tdiode2 = t2;
	
	//sound code  :)
	TransferSound *snd = IPC->soundData;
	IPC->soundData = 0;

	if (0 != snd) {

		for (i=0; i<snd->count; i++) {
			s32 chan = getFreeSoundChannel();

			if (chan >= 0) {
				startSound(snd->data[i].rate, snd->data[i].data, snd->data[i].len, chan, snd->data[i].vol, snd->data[i].pan, snd->data[i].format);
			}
		}
	}
	
	
	//VBLANK_INTR_WAIT_FLAGS|=flags;
	//REG_IF=flags;
}

//change power management register
void writePM(int flags) {
	call3 spi_write=(call3)0x2369;
	spi_write(SPI_DEVICE_POWER,0,1);
	spi_write(SPI_DEVICE_POWER,flags,0);
}

//shut down, wait for lid to open
void arm7sleep() {
	u32 ie_save;
	writePM(PM_TOPLIGHT | PM_BOTTOMLIGHT | PM_SOUND | PM_LED_SLOW);
	swiChangeSoundBias(0,0x400);
	ie_save=REG_IE;
	REG_IE=IRQ_LID;
	swiSleep();
	REG_IF=~0;
	REG_IE=ie_save;
	writePM(PM_TOPLIGHT | PM_BOTTOMLIGHT | PM_SOUND);
	swiChangeSoundBias(1,0x400);
}

static void clearmem(u32 dest, u32 size) {
	int i=0;
	swiFastCopy(&i, (u32*)dest, 0x01000000 | (size>>2));
}

void hardware_init() {
	REG_IME=REG_IE=0;
	
	writePM(PM_TOPLIGHT | PM_BOTTOMLIGHT | PM_SOUND);
	POWER_CR=POWER_SOUND;
	swiChangeSoundBias(1,0x400);

	//reset DMA
	clearmem(0x40000B0, 0x30);

	//((call0)0x119d)();
}


int main() {
	u32 i,top_light,bottom_light;

	IPC_ARM9=-1;
	IPC_XKEYS=0;
	top_light=1;
	bottom_light=1;
	
	hardware_init();
	

	DISP_SR=DISP_VBLANK_IRQ;
	/*IRQ_HANDLER=interrupthandler;
	REG_IE=IRQ_VBLANK;
	REG_IF=~0;
	REG_IME=IME_ENABLE;*/
	
	irqInit();
	irqSet(IRQ_VBLANK, interrupthandler);
	irqEnable(IRQ_VBLANK);
	

	load_PersonalData();
	load_WifiData(0,(u32*)WIFI_SETTING);
	
	touchInit=false;
	SOUND_CR=SOUND_ENABLE|SOUND_VOL(127); //Onscreen keyboard needs this.
	
	while(1) {
		swiWaitForVBlank();

		IPC_XKEYS=REG_KEYXY;
		
		i=IPC_ARM9;
		if(i!=-1) {
			IPC_ARM9=-1;
			switch(i) {
			
				case IPC_NDS:
					REG_IME=IME_DISABLE;
					if(load_card())
					{
						hardware_init();
						IPC_ARM7=IPC_NDS_GO;
						writePM((top_light*PM_TOPLIGHT) | (bottom_light*PM_BOTTOMLIGHT) | PM_SOUND);
								//So the backlight is off for gameplay, if user requests this.
						(*(call0*)0x27ffe34)();
					}
					else
					{
						REG_IME=IME_ENABLE;
						IPC_ARM7=IPC_NDS_STOP;
						break;
					}
					
				case IPC_PASSME:
					hardware_init();
					((call0)0x80000c0)();
					
				case IPC_GBA:
					if(PersonalData->gbaScreen)
						i=(bottom_light*PM_BOTTOMLIGHT) | PM_SOUND;
					else
						i=(top_light*PM_TOPLIGHT) | PM_SOUND;
					writePM(i);
					swiChangeSoundBias(0,0x400);
					((call0)0x11a4)();
					
				case IPC_LIDCLOSE:
					writePM(PM_SOUND);
					break;
					
				case IPC_LIDOPEN:
					writePM((top_light*PM_TOPLIGHT) | (bottom_light*PM_BOTTOMLIGHT) | PM_SOUND);
					break;
					
				case IPC_SLEEP:
					arm7sleep();
					break;
					
				case IPC_CALL_ARM7:
					((call0)IPC_ARM9_P1)();
					break;
					
				case IPC_WRITE_PSETTINGS:
					save_PersonalData();
					break;
					
				case IPC_POWEROFF:
					writePM(PM_OFF);
					break;
					
				case IPC_SET_TOP_LIGHT:
					top_light=!top_light;
					writePM((top_light*PM_TOPLIGHT) | (bottom_light*PM_BOTTOMLIGHT) | PM_SOUND);
					break;
					
				case IPC_SET_BOTTOM_LIGHT:
					bottom_light=!bottom_light;
					writePM((top_light*PM_TOPLIGHT) | (bottom_light*PM_BOTTOMLIGHT) | PM_SOUND);
					break;

				case IPC_ALARM:
					SCHANNEL_TIMER(8)=SOUND_FREQ((440*8));
					SCHANNEL_CR(8)=SCHANNEL_ENABLE|SOUND_FORMAT_PSG|SOUND_VOL(127);
					break;

				case IPC_ALARMOFF:
					SCHANNEL_CR(8)=0;
					break;
					
				case IPC_WRITE_RTC:
					REG_IME = IME_DISABLE;
					integerToBCD((u8*)(IPC_RTC_BASE+1),7);
					SETRTC((u8*)(IPC_RTC_BASE+0));
					REG_IME = IME_ENABLE;
					break;
					
				case IPC_CAL_TOUCHSCREEN:
					touchInit = false;
					break;
					
				case IPC_WRITE_WIFI_0:
					save_WifiData(0,(u32*)WIFI_SETTING);
					break;
				case IPC_WRITE_WIFI_1:
					save_WifiData(1,(u32*)(WIFI_SETTING+0x100));
					break;
				case IPC_WRITE_WIFI_2:
					save_WifiData(2,(u32*)(WIFI_SETTING+0x200));
					break;
					
			}
		}
	}
	return 0;
}

